#! /usr/bin/env python
#
# Python module for TOTP
# Alejandro Perez Mendez <alex.perez-mendez@jisc.ac.uk>
#
# Most of the TOTP code has been borrowed from https://github.com/pyotp/pyotp

"""
This module implements support for TOTP 2-factor authentication, assuming
end users use a concatenation of "password+code" as their passwords.

The module works as follows:

If a request is processed and there is an TOTP secret configured for the user,
then one of the following is applied:
- If there is a Cleartext-Password configured for the user, its value for this
  request is modified by appending the expected TOTP code to it.
- If there is a User-Password attribute in the request, the last characters
  are assumed to be the TOTP code. If the TOTP code is good, the value
  of User-Password is modified by removing the characters corresponding to
  the TOTP code.

In both cases, after having changed the relevant bits, the module lets
FreeRADIUS to continue with the processing of the request.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import base64
import datetime
import hashlib
import hmac
import random
import sys
import time
import traceback

# Some configuration for TOTP
INTERVAL = 30
DIGITS = 6


class TOTP(object):
    """
    TOTP handler (mostly borrowed from https://github.com/pyotp/pyotp)
    """
    def __init__(self, s, interval=30, digits=6, digest=hashlib.sha1):
        self.digits = digits
        self.digest = digest
        self.secret = s
        self.interval = interval

    def generate_otp(self, input):
        if input < 0:
            raise ValueError('input must be positive integer')
        hasher = hmac.new(self.byte_secret(), self.int_to_bytestring(input), self.digest)
        hmac_hash = bytearray(hasher.digest())
        offset = hmac_hash[-1] & 0xf
        code = ((hmac_hash[offset] & 0x7f) << 24 |
                (hmac_hash[offset + 1] & 0xff) << 16 |
                (hmac_hash[offset + 2] & 0xff) << 8 |
                (hmac_hash[offset + 3] & 0xff))
        str_code = str(code % 10 ** self.digits)
        while len(str_code) < self.digits:
            str_code = '0' + str_code

        return str_code

    def byte_secret(self):
        missing_padding = len(self.secret) % 8
        if missing_padding != 0:
            self.secret += '=' * (8 - missing_padding)
        return base64.b32decode(self.secret, casefold=True)

    @staticmethod
    def int_to_bytestring(i, padding=8):
        result = bytearray()
        while i != 0:
            result.append(i & 0xFF)
            i >>= 8
        # It's necessary to convert the final result from bytearray to bytes
        # because the hmac functions in python 2.6 and 3.3 don't work with
        # bytearray
        return bytes(bytearray(reversed(result)).rjust(padding, b'\0'))

    def now(self):
        return self.generate_otp(self.timecode(datetime.datetime.now()))

    def next(self):
        return self.generate_otp(self.timecode(datetime.datetime.now()) + 1)

    def prev(self):
        return self.generate_otp(self.timecode(datetime.datetime.now()) - 1)

    def timecode(self, for_time):
        i = time.mktime(for_time.timetuple())
        return int(i / self.interval)

    @staticmethod
    def random_base32(length=32):
        rnd = random.SystemRandom()
        chars = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567')
        return ''.join(rnd.choice(chars) for _ in range(length))

    def build_uri(self,     label):
        """
        Returns the provisioning URI for the OTP; works for either TOTP or HOTP.
        This can then be encoded in a QR Code and used to provision the Google
        Authenticator app.
        """
        return 'otpauth://totp/{0}?secret={1}&digits={2}&period={3}'.format(label, self.secret, self.digits, self.interval)


def get_attr(p, attr_name):
    """
    Gets the first occurence of a RADIUS attribute in a tuple of tuples
    """
    try:
        return next(value for (name, value) in p if name == attr_name)
    except StopIteration:
        return None


def authorize(p):
    # This needs to be imported here so the module can be used outside FreeRADIUS for secret generation
    import radiusd
    """
    Performs the "authorize" functionality.
    """
    try:
        user_password = get_attr(p["request"], 'User-Password')
        cleartext_password = get_attr(p["config"], 'Cleartext-Password')
        secret = get_attr(p["config"], 'Moonshot-OTP-Secret')
        if not secret:
            radiusd.radlog(radiusd.L_WARN, 'OTP shared secret was not found for this user')
            return radiusd.RLM_MODULE_NOOP

        totp = TOTP(secret, interval=INTERVAL, digits=DIGITS)

        # If the request contains a User-Password, then we take the last DIGITS digits
        # as the OTP code, verify it and rewrite the password without the code
        if user_password:
            otp_code = user_password[-DIGITS:]
            actual_password = user_password[:-DIGITS]
            # Check OTP code
            for code in (totp.prev(), totp.now(), totp.next()):
                if (otp_code == code):
                    # rewrite password
                    request = ( (str('User-Password'), str(':='), actual_password), )
                    return radiusd.RLM_MODULE_OK, {"request": request}

            radiusd.radlog(radiusd.L_ERR, "Provided OTP code differs from the expected one")
            return radiusd.RLM_MODULE_FAIL

        # If there is a Cleartext-Password in the config, then we just append the expected
        # OTP code to it
        elif cleartext_password:
            code = str(totp.now())
            config = ( (str('Cleartext-Password'), str(':='), cleartext_password + code), )
            return radiusd.RLM_MODULE_OK, {"config": config}

        # if none of these conditions match, module fails
        else:
            radiusd.radlog(radiusd.L_ERR, 'OTP requires config:Cleartext-Password or request:User-Password')
            return radiusd.RLM_MODULE_FAIL

    except Exception as ex:
        radiusd.radlog(radiusd.L_ERR, traceback.format_exc())
        return radiusd.RLM_MODULE_FAIL


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("You must provide a label for this account as the first argument")
        sys.exit(1)
    secret = TOTP.random_base32()
    print("Secret: {0}".format(secret))
    print("URL: {0}".format(TOTP(secret).build_uri(sys.argv[1])))
